from django import template

register = template.Library()


@register.filter
def power10(value, k=0):
    """converts the value to a power of 10 representation"""
    try:
        return float(value) / (10 ** int(k))
    except (ValueError, TypeError):
        return None


@register.filter
def formatnumber(value, p=1):
    """return a formated number with thousands seperators"""
    try:
        return "{:,.{}f}".format(float(value), int(p))
    except (ValueError, TypeError):
        return None



# Generated by Django 2.2.8 on 2019-12-14 17:12

from django.db import migrations


class Migration(migrations.Migration):

    atomic = False

    dependencies = [
        ("freight", "0013_auto_20191214_1522"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="EveOrganization",
            new_name="EveEntity",
        ),
    ]



# Generated by Django 3.1.9 on 2021-05-06 15:44

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("authentication", "0017_remove_fleetup_permission"),
        ("eveonline", "0014_auto_20210105_1413"),
        ("freight", "0016_default_pricing"),
    ]

    operations = [
        migrations.AlterField(
            model_name="contract",
            name="acceptor",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="character of acceptor or None if accepted by corp",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_acceptor",
                to="eveonline.evecharacter",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="acceptor_corporation",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="corporation of acceptor",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_acceptor_corporation",
                to="eveonline.evecorporationinfo",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="date_notified",
            field=models.DateTimeField(
                blank=True,
                db_index=True,
                default=None,
                help_text="datetime of latest notification, None = none has been sent",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="end_location",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_end_location",
                to="freight.location",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="handler",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts",
                to="freight.contracthandler",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="issuer",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_issuer",
                to="eveonline.evecharacter",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="issuer_corporation",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_issuer_corporation",
                to="eveonline.evecorporationinfo",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="pricing",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="contracts",
                to="freight.pricing",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="start_location",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_start_location",
                to="freight.location",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="status",
            field=models.CharField(
                choices=[
                    ("outstanding", "outstanding"),
                    ("in_progress", "in progress"),
                    ("finished_issuer", "finished issuer"),
                    ("finished_contractor", "finished contractor"),
                    ("finished", "finished"),
                    ("canceled", "canceled"),
                    ("rejected", "rejected"),
                    ("failed", "failed"),
                    ("deleted", "deleted"),
                    ("reversed", "reversed"),
                ],
                db_index=True,
                max_length=32,
            ),
        ),
        migrations.AlterField(
            model_name="contractcustomernotification",
            name="contract",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="customer_notifications",
                to="freight.contract",
            ),
        ),
        migrations.AlterField(
            model_name="contractcustomernotification",
            name="status",
            field=models.CharField(
                choices=[
                    ("outstanding", "outstanding"),
                    ("in_progress", "in progress"),
                    ("finished_issuer", "finished issuer"),
                    ("finished_contractor", "finished contractor"),
                    ("finished", "finished"),
                    ("canceled", "canceled"),
                    ("rejected", "rejected"),
                    ("failed", "failed"),
                    ("deleted", "deleted"),
                    ("reversed", "reversed"),
                ],
                db_index=True,
                max_length=32,
            ),
        ),
        migrations.AlterField(
            model_name="contracthandler",
            name="character",
            field=models.ForeignKey(
                default=None,
                help_text="character used for syncing contracts",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="authentication.characterownership",
            ),
        ),
        migrations.AlterField(
            model_name="location",
            name="name",
            field=models.CharField(
                db_index=True,
                help_text="In-game name of this station or structure",
                max_length=100,
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="end_location",
            field=models.ForeignKey(
                help_text="Destination station or structure for courier route",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="freight.location",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="start_location",
            field=models.ForeignKey(
                help_text="Starting station or structure for courier route",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="freight.location",
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2019-10-28 10:36

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("freight", "0003_auto_20191028_1023"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="contract",
            name="price",
        ),
    ]



# Generated by Django 2.2.5 on 2019-10-28 10:23

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("freight", "0002_auto_20191015_1229"),
    ]

    operations = [
        migrations.AlterField(
            model_name="contract",
            name="acceptor",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contract_acceptor",
                to="eveonline.EveCharacter",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="date_accepted",
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="contract",
            name="date_completed",
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name="contract",
            name="date_notified",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="datetime of latest notification, None = none has been sent",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="issues",
            field=models.TextField(
                blank=True,
                default=None,
                help_text="List or price check issues as JSON array of strings or None",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="pricing",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                to="freight.Pricing",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="title",
            field=models.CharField(blank=True, default=None, max_length=100, null=True),
        ),
    ]



from django.db.models.signals import post_save
from django.dispatch import receiver

from .models import Pricing
from .tasks import update_contracts_pricing


@receiver(post_save, sender=Pricing, dispatch_uid="id_update_contracts_pricing")
def pricing_save_handler(sender, instance, *args, **kwargs):
    """contract pricing needs to be updated after every pricing change"""
    update_contracts_pricing.delay()



from django.conf.urls import url

from . import views

app_name = "freight"

urlpatterns = [
    url(r"^$", views.index, name="index"),
    url(
        r"^setup_contract_handler/$",
        views.setup_contract_handler,
        name="setup_contract_handler",
    ),
    url(r"^add_location/$", views.add_location, name="add_location"),
    url(r"^add_location_2/$", views.add_location_2, name="add_location_2"),
    url(r"^calculator/$", views.calculator, name="calculator"),
    url(r"^calculator/(?P<pricing_pk>[0-9]+)/$", views.calculator, name="calculator"),
    url(r"^contract_list_all/$", views.contract_list_all, name="contract_list_all"),
    url(r"^contract_list_user/$", views.contract_list_user, name="contract_list_user"),
    url(
        r"^contract_list_data/(?P<category>.+)/$",
        views.contract_list_data,
        name="contract_list_data",
    ),
    url(r"^statistics/$", views.statistics, name="statistics"),
    url(
        r"^statistics_routes_data/$",
        views.statistics_routes_data,
        name="statistics_routes_data",
    ),
    url(
        r"^statistics_pilots_data/$",
        views.statistics_pilots_data,
        name="statistics_pilots_data",
    ),
    url(
        r"^statistics_pilot_corporations_data/$",
        views.statistics_pilot_corporations_data,
        name="statistics_pilot_corporations_data",
    ),
    url(
        r"^statistics_customer_data/$",
        views.statistics_customer_data,
        name="statistics_customer_data",
    ),
]



import math

from django import forms
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator

from .models import Pricing


class CalculatorForm(forms.Form):
    pricing = forms.ModelChoiceField(
        queryset=Pricing.objects.select_related("start_location", "end_location")
        .filter(is_active=True)
        .order_by("start_location__name", "end_location__name"),
        label="Route",
        help_text="Pick a route for your courier contract",
        empty_label=None,
    )
    volume = forms.IntegerField(
        help_text="Est. volume of your cargo in m3",
        required=False,
        validators=[MinValueValidator(0)],
    )
    collateral = forms.IntegerField(
        help_text=(
            "Collaterial in ISK, must be roughly equal to the est. "
            "value of your cargo"
        ),
        required=False,
        validators=[MinValueValidator(0)],
    )

    def clean(self):
        pricing = self.cleaned_data["pricing"]
        issue_prefix = "⚠ Issues:"

        if pricing.requires_volume() and self.cleaned_data["volume"] is None:
            raise ValidationError("{} volume is required".format(issue_prefix))

        if pricing.requires_collateral() and self.cleaned_data["collateral"] is None:
            raise ValidationError("{} collateral is required".format(issue_prefix))

        volume = self.cleaned_data["volume"]
        collateral = self.cleaned_data["collateral"]
        issues = pricing.get_contract_price_check_issues(volume, collateral)

        if issues:
            raise ValidationError("{} {}".format(issue_prefix, ", ".join(issues)))

    def get_calculated_data(self, pricing: object) -> tuple:
        if self.is_valid():
            if self.cleaned_data["volume"]:
                volume = int(self.cleaned_data["volume"])
            else:
                volume = 0
            if self.cleaned_data["collateral"]:
                collateral = int(self.cleaned_data["collateral"])
            else:
                collateral = 0
            price = (
                math.ceil(pricing.get_calculated_price(volume, collateral) / 1000000)
                * 1000000
            )

        else:
            volume = None
            collateral = None
            price = None

        return volume, collateral, price


class AddLocationForm(forms.Form):
    location_id = forms.IntegerField(
        label="Location ID",
        help_text="Eve Online ID for a station or an Upwell structure",
    )



import datetime

from django.conf import settings
from django.contrib.auth.decorators import login_required, permission_required
from django.db import models
from django.db.models import Count, Q, Sum
from django.forms import HiddenInput
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.utils.html import format_html, mark_safe
from django.utils.timezone import now
from esi.decorators import token_required
from esi.models import Token

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from app_utils.messages import messages_plus

from . import __title__, tasks
from .app_settings import (
    FREIGHT_APP_NAME,
    FREIGHT_OPERATION_MODE,
    FREIGHT_STATISTICS_MAX_DAYS,
)
from .models import Contract, ContractHandler, EveEntity, Freight, Location, Pricing

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

ADD_LOCATION_TOKEN_TAG = "freight_add_location_token"
CONTRACT_LIST_USER = "user"
CONTRACT_LIST_ACTIVE = "active"
CONTRACT_LIST_ALL = "all"


def add_common_context(request, context: dict) -> dict:
    """adds the common context used by all view"""
    pending_user_count = (
        Contract.objects.all().issued_by_user(request.user).pending_count()
    )
    operation_mode = Freight.operation_mode_friendly(FREIGHT_OPERATION_MODE)
    new_context = {
        **{
            "app_title": FREIGHT_APP_NAME,
            "pending_all_count": Contract.objects.all().pending_count(),
            "pending_user_count": pending_user_count,
            "setup_contract_handler_label": f"Setup {operation_mode}",
        },
        **context,
    }
    return new_context


@login_required
@permission_required("freight.basic_access")
def index(request):
    return redirect("freight:calculator")


@login_required
@permission_required("freight.use_calculator")
def contract_list_user(request):
    try:
        user_name = request.user.profile.main_character.character_name
    except AttributeError:
        user_name = request.user.user_name

    context = {
        "page_title": "My Contracts",
        "category": CONTRACT_LIST_USER,
        "user_name": user_name,
    }
    return render(
        request, "freight/contracts_user.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.view_contracts")
def contract_list_all(request):
    context = {
        "page_title": "All Contracts",
        "category": CONTRACT_LIST_ALL,
    }
    return render(
        request, "freight/contracts_all.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.basic_access")
def contract_list_data(request, category) -> JsonResponse:
    """Return list of outstanding contracts for contract_list AJAX call."""

    def character_format(obj):
        return obj.character_name if obj else None

    contracts_data = list()
    for contract in _contracts_for_contract_list(category, request):
        if contract.has_pricing:
            route_name = contract.pricing.name
            if not contract.has_pricing_errors:
                tooltip_text = route_name
                icon_html = format_html(
                    '<span class="{}"><i class="fas fa-check" title="{}"></i></span>',
                    "text-success",
                    tooltip_text,
                )
            else:
                tooltip_text = "{}\n{}".format(
                    route_name, "\n".join(contract.get_issue_list())
                )
                icon_html = format_html(
                    (
                        '<span class="{}">'
                        '<i class="fas fa-exclamation-triangle" title="{}"></i>'
                        "</span>"
                    ),
                    "text-danger",
                    tooltip_text,
                )

            pricing_check = icon_html
        else:
            route_name = ""
            pricing_check = "-"

        if contract.title or settings.DEBUG:
            if settings.DEBUG:
                title = "{}{}".format(
                    f"{contract.title} " if contract.title else "", contract.contract_id
                )
            else:
                title = contract.title

            notes = format_html('<i class="far fa-envelope" title="{}"></i>', title)
        else:
            notes = ""

        start_location_html = format_html(
            '<span class="dotted-underline" title="{}">{}</span> {}',
            contract.start_location,
            contract.start_location.solar_system_name,
            notes,
        )
        end_location_html = format_html(
            '<span class="dotted-underline" title="{}">{}</span>',
            contract.end_location,
            contract.end_location.solar_system_name,
        )
        contracts_data.append(
            {
                "contract_id": contract.contract_id,
                "status": str(contract.status),
                "start_location": {
                    "display": start_location_html,
                    "sort": contract.start_location.name,
                },
                "end_location": {
                    "display": end_location_html,
                    "sort": contract.end_location.name,
                },
                "reward": contract.reward,
                "collateral": contract.collateral,
                "volume": contract.volume,
                "date_issued": contract.date_issued.isoformat(),
                "date_expired": contract.date_expired.isoformat(),
                "issuer": character_format(contract.issuer),
                "date_accepted": contract.date_accepted.isoformat()
                if contract.date_accepted
                else None,
                "acceptor": contract.acceptor_name,
                "has_pricing": contract.has_pricing,
                "has_pricing_errors": contract.has_pricing_errors,
                "pricing_check": pricing_check,
                "route_name": route_name,
                "is_in_progress": contract.is_in_progress,
                "is_failed": contract.is_failed,
                "is_completed": contract.is_completed,
            }
        )
    return JsonResponse(contracts_data, safe=False)


def _contracts_for_contract_list(category, request) -> models.QuerySet:
    """returns contracts corresponding to given category"""
    contracts_qs = Contract.objects.select_related(
        "acceptor",
        "acceptor_corporation",
        "end_location",
        "issuer",
        "start_location",
        "pricing",
        "pricing__start_location",
        "pricing__end_location",
    )
    if category == CONTRACT_LIST_ACTIVE:
        if not request.user.has_perm("freight.view_contracts"):
            return contracts_qs.none()
        return contracts_qs.filter(
            status__in=[
                Contract.Status.OUTSTANDING,
                Contract.Status.IN_PROGRESS,
            ]
        ).exclude(date_expired__lt=now())
    elif category == CONTRACT_LIST_ALL:
        if not request.user.has_perm("freight.view_contracts"):
            return contracts_qs.none()
        return contracts_qs
    elif category == CONTRACT_LIST_USER:
        if not request.user.has_perm("freight.use_calculator"):
            return contracts_qs.none()
        return contracts_qs.issued_by_user(user=request.user).filter(
            status__in=[
                Contract.Status.OUTSTANDING,
                Contract.Status.IN_PROGRESS,
                Contract.Status.FINISHED,
                Contract.Status.FAILED,
            ]
        )
    raise ValueError("Invalid category: {}".format(category))


@login_required
@permission_required("freight.use_calculator")
def calculator(request, pricing_pk=None):
    from .forms import CalculatorForm

    if request.method != "POST":
        pricing = Pricing.objects.get_or_default(pricing_pk)
        form = CalculatorForm(initial={"pricing": pricing})
        price = None
        volume = None
        collateral = None
    else:
        form = CalculatorForm(request.POST)
        request.POST._mutable = True
        pricing = Pricing.objects.get_or_default(form.data["pricing"])
        volume, collateral, price = form.get_calculated_data(pricing)

    if pricing:
        price_per_volume_eff = pricing.price_per_volume_eff()
        if not pricing.requires_volume():
            form.fields["volume"].widget = HiddenInput()
        if not pricing.requires_collateral():
            form.fields["collateral"].widget = HiddenInput()
    else:
        price_per_volume_eff = None

    if price:
        if pricing.days_to_expire:
            expires_on = datetime.datetime.now(
                datetime.timezone.utc
            ) + datetime.timedelta(days=pricing.days_to_expire)
        else:
            expires_on = None
    else:
        collateral = None
        volume = None
        expires_on = None

    handler = ContractHandler.objects.select_related("organization").first()
    if handler:
        organization_name = handler.organization.name
        availability = handler.get_availability_text_for_contracts()
    else:
        organization_name = None
        availability = None

    context = {
        "page_title": "Reward Calculator",
        "form": form,
        "pricing": pricing,
        "price": price,
        "organization_name": organization_name,
        "collateral": collateral if collateral is not None else 0,
        "volume": volume if volume is not None else None,
        "expires_on": expires_on,
        "availability": availability,
        "pricing_price_per_volume_eff": price_per_volume_eff,
    }
    return render(
        request, "freight/calculator.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.setup_contract_handler")
@token_required(scopes=ContractHandler.get_esi_scopes())
def setup_contract_handler(request, token):
    success = True
    token_char = EveCharacter.objects.get(character_id=token.character_id)

    if (
        EveEntity.get_category_for_operation_mode(FREIGHT_OPERATION_MODE)
        == EveEntity.Category.ALLIANCE
    ) and token_char.alliance_id is None:
        messages_plus.error(
            request,
            "Can not setup contract handler, "
            "because {} is not a member of any alliance".format(token_char),
        )
        success = False

    owned_char = None
    if success:
        try:
            owned_char = CharacterOwnership.objects.get(
                user=request.user, character=token_char
            )
        except CharacterOwnership.DoesNotExist:
            messages_plus.error(
                request,
                mark_safe(
                    "You can only use your main or alt characters to setup "
                    " the contract handler. "
                    "However, character <strong>{}</strong> is neither. ".format(
                        token_char.character_name
                    )
                ),
            )
            success = False

    if success:
        handler = ContractHandler.objects.first()
        if handler and handler.operation_mode != FREIGHT_OPERATION_MODE:
            messages_plus.error(
                request,
                "There already is a contract handler installed for a "
                "different operation mode. You need to first delete the "
                "existing contract handler in the admin section "
                "before you can set up this app for a different operation mode.",
            )
            success = False

    if success:
        organization, _ = EveEntity.objects.update_or_create_from_evecharacter(
            token_char,
            EveEntity.get_category_for_operation_mode(FREIGHT_OPERATION_MODE),
        )

        handler, _ = ContractHandler.objects.update_or_create(
            organization=organization,
            defaults={
                "character": owned_char,
                "operation_mode": FREIGHT_OPERATION_MODE,
            },
        )
        tasks.run_contracts_sync.delay(force_sync=True, user_pk=request.user.pk)
        messages_plus.success(
            request,
            mark_safe(
                "Contract Handler setup started for "
                "<strong>{}</strong> organization "
                "with <strong>{}</strong> as sync character. "
                "Operation mode: <strong>{}</strong>. "
                "Started syncing of courier contracts. "
                "You will receive a report once it is completed.".format(
                    organization.name,
                    handler.character.character.character_name,
                    handler.operation_mode_friendly,
                )
            ),
        )

    return redirect("freight:index")


@login_required
@token_required(scopes=Location.get_esi_scopes())
@permission_required("freight.add_location")
def add_location(request, token):
    request.session[ADD_LOCATION_TOKEN_TAG] = token.pk
    return redirect("freight:add_location_2")


@login_required
@permission_required("freight.add_location")
def add_location_2(request):
    from .forms import AddLocationForm

    if ADD_LOCATION_TOKEN_TAG not in request.session:
        raise RuntimeError("Missing token in session")
    else:
        token = Token.objects.get(pk=request.session[ADD_LOCATION_TOKEN_TAG])

    if request.method != "POST":
        form = AddLocationForm()

    else:
        form = AddLocationForm(request.POST)
        if form.is_valid():
            location_id = form.cleaned_data["location_id"]
            try:
                location, created = Location.objects.update_or_create_from_esi(
                    token=token, location_id=location_id, add_unknown=False
                )
                action_txt = "Added:" if created else "Updated:"
                messages_plus.success(
                    request,
                    mark_safe(
                        "{} <strong>{}</strong>".format(action_txt, location.name)
                    ),
                )
                return redirect("freight:add_location_2")

            except Exception as ex:
                messages_plus.error(
                    request,
                    "Failed to add location with token from {} "
                    "for location ID {}: {}".format(
                        token.character_name, location_id, type(ex).__name__
                    ),
                )

    context = {
        "page_title": "Add / Update Location",
        "form": form,
        "token_char_name": token.character_name,
    }
    return render(
        request, "freight/add_location.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.view_statistics")
def statistics(request):
    context = {
        "page_title": "Statistics",
        "max_days": FREIGHT_STATISTICS_MAX_DAYS,
    }
    return render(
        request, "freight/statistics.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.view_statistics")
def statistics_routes_data(request):
    """returns totals for statistics as JSON"""

    cutoff_date = now() - datetime.timedelta(days=FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(contracts__status=Contract.Status.FINISHED) & Q(
        contracts__date_completed__gte=cutoff_date
    )
    route_totals = (
        Pricing.objects.annotate(
            contracts_count=Count("contracts", filter=finished_contracts)
        )
        .select_related("start_location", "end_location")
        .annotate(rewards=Sum("contracts__reward", filter=finished_contracts))
        .annotate(collaterals=Sum("contracts__collateral", filter=finished_contracts))
        .annotate(volume=Sum("contracts__volume", filter=finished_contracts))
        .annotate(
            pilots=Count(
                "contracts__acceptor", distinct=True, filter=finished_contracts
            )
        )
        .annotate(
            customers=Count(
                "contracts__issuer", distinct=True, filter=finished_contracts
            )
        )
    )

    totals = list()
    for route in route_totals:
        if route.contracts_count > 0:
            totals.append(
                {
                    "name": route.name,
                    "contracts": route.contracts_count,
                    "rewards": route.rewards,
                    "collaterals": route.collaterals,
                    "volume": route.volume,
                    "pilots": route.pilots,
                    "customers": route.customers,
                }
            )

    return JsonResponse(totals, safe=False)


@login_required
@permission_required("freight.view_statistics")
def statistics_pilots_data(request):
    """returns totals for statistics as JSON"""

    cutoff_date = now() - datetime.timedelta(days=FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(contracts_acceptor__status=Contract.Status.FINISHED) & Q(
        contracts_acceptor__date_completed__gte=cutoff_date
    )

    pilot_totals = (
        EveCharacter.objects.exclude(contracts_acceptor__isnull=True)
        .annotate(
            contracts_count=Count("contracts_acceptor", filter=finished_contracts)
        )
        .annotate(rewards=Sum("contracts_acceptor__reward", filter=finished_contracts))
        .annotate(
            collaterals=Sum("contracts_acceptor__collateral", filter=finished_contracts)
        )
        .annotate(volume=Sum("contracts_acceptor__volume", filter=finished_contracts))
    )

    totals = list()
    for pilot in pilot_totals:
        if pilot.contracts_count > 0:
            totals.append(
                {
                    "name": pilot.character_name,
                    "corporation": pilot.corporation_name,
                    "contracts": pilot.contracts_count,
                    "rewards": pilot.rewards,
                    "collaterals": pilot.collaterals,
                    "volume": pilot.volume,
                }
            )

    return JsonResponse(totals, safe=False)


@login_required
@permission_required("freight.view_statistics")
def statistics_pilot_corporations_data(request):
    """returns totals for statistics as JSON"""

    cutoff_date = now() - datetime.timedelta(days=FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(
        contracts_acceptor_corporation__status=Contract.Status.FINISHED
    ) & Q(contracts_acceptor_corporation__date_completed__gte=cutoff_date)

    corporation_totals = (
        EveCorporationInfo.objects.exclude(contracts_acceptor_corporation__isnull=True)
        .select_related("alliance")
        .annotate(
            contracts_count=Count(
                "contracts_acceptor_corporation", filter=finished_contracts
            )
        )
        .annotate(
            rewards=Sum(
                "contracts_acceptor_corporation__reward", filter=finished_contracts
            )
        )
        .annotate(
            collaterals=Sum(
                "contracts_acceptor_corporation__collateral", filter=finished_contracts
            )
        )
        .annotate(
            volume=Sum(
                "contracts_acceptor_corporation__volume", filter=finished_contracts
            )
        )
    )

    totals = list()
    for corporation in corporation_totals:
        if corporation.contracts_count > 0:
            alliance = (
                corporation.alliance.alliance_name if corporation.alliance else ""
            )
            totals.append(
                {
                    "name": corporation.corporation_name,
                    "alliance": alliance,
                    "contracts": corporation.contracts_count,
                    "rewards": corporation.rewards,
                    "collaterals": corporation.collaterals,
                    "volume": corporation.volume,
                }
            )

    return JsonResponse(totals, safe=False)


@login_required
@permission_required("freight.view_statistics")
def statistics_customer_data(request):
    """returns totals for statistics as JSON"""

    cutoff_date = now() - datetime.timedelta(days=FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(contracts_issuer__status=Contract.Status.FINISHED) & Q(
        contracts_issuer__date_completed__gte=cutoff_date
    )
    customer_totals = (
        EveCharacter.objects.exclude(contracts_issuer__isnull=True)
        .annotate(contracts_count=Count("contracts_issuer", filter=finished_contracts))
        .annotate(rewards=Sum("contracts_issuer__reward", filter=finished_contracts))
        .annotate(
            collaterals=Sum("contracts_issuer__collateral", filter=finished_contracts)
        )
        .annotate(volume=Sum("contracts_issuer__volume", filter=finished_contracts))
    )

    totals = list()
    for customer in customer_totals:
        if customer.contracts_count > 0:
            totals.append(
                {
                    "name": customer.character_name,
                    "corporation": customer.corporation_name,
                    "contracts": customer.contracts_count,
                    "rewards": customer.rewards,
                    "collaterals": customer.collaterals,
                    "volume": customer.volume,
                }
            )

    return JsonResponse(totals, safe=False)



# Generated by Django 2.2.9 on 2020-01-30 23:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("freight", "0014_auto_20191214_1712"),
    ]

    operations = [
        migrations.RenameField(
            model_name="pricing",
            old_name="active",
            new_name="is_active",
        ),
        migrations.AddField(
            model_name="pricing",
            name="is_bidirectional",
            field=models.BooleanField(
                default=True,
                help_text="Whether this pricing is valid for contracts in either direction or only the one specified",
            ),
        ),
    ]



from celery import chain, shared_task

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .models import Contract, ContractHandler, Location

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def _get_user(user_pk) -> User:
    """returns the user or None. Logs if user is requested but can't be found."""
    user = None
    if user_pk:
        try:
            user = User.objects.get(pk=user_pk)
        except User.DoesNotExist:
            logger.warning("Ignoring non-existing user with pk {}".format(user_pk))
    return user


def _get_contract_handler() -> ContractHandler:
    handler = ContractHandler.objects.first()
    if not handler:
        logger.warning("No contract handler was found")
        raise ObjectDoesNotExist()
    else:
        return handler


@shared_task
def update_contracts_esi(force_sync=False, user_pk=None) -> None:
    """start syncing contracts"""
    _get_contract_handler().update_contracts_esi(force_sync, user=_get_user(user_pk))


@shared_task
def send_contract_notifications(force_sent=False, rate_limted=True) -> None:
    """Send notification about outstanding contracts that have pricing"""
    try:
        Contract.objects.send_notifications(force_sent, rate_limted)

    except Exception as ex:
        logger.exception("An unexpected error ocurred: {}".format(ex))


@shared_task
def run_contracts_sync(force_sync=False, user_pk=None) -> None:
    """main task coordinating contract sync"""
    my_chain = chain(
        update_contracts_esi.si(force_sync, user_pk), send_contract_notifications.si()
    )
    my_chain.delay()


@shared_task
def update_contracts_pricing() -> None:
    """Updates pricing for all contracts"""
    logger.info("Started updating contracts pricing")
    try:
        Contract.objects.update_pricing()

    except Exception as ex:
        logger.exception("An unexpected error ocurred: {}".format(ex))


@shared_task
def update_location(location_id: int) -> None:
    """Updates the location from ESI"""
    try:
        Location.objects.get(id=location_id)
    except Location.DoesNotExist:
        logger.warning(
            "Tried to update a non-existing location with ID {}".format(location_id)
        )
    else:
        token = _get_contract_handler().token()
        Location.objects.update_or_create_from_esi(location_id=location_id, token=token)


@shared_task
def update_locations(location_ids: list) -> None:
    """Updates the locations from ESI"""

    for location_id in location_ids:
        update_location.delay(location_id)



# Generated by Django 2.2.8 on 2019-12-14 15:22

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveonline", "0010_alliance_ticker"),
        ("freight", "0012_auto_20191210_1406"),
    ]

    operations = [
        migrations.AddField(
            model_name="contract",
            name="acceptor_corporation",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="corporation of acceptor",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contract_acceptor_corporation",
                to="eveonline.EveCorporationInfo",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="acceptor",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="character of acceptor or None if accepted by corp",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contract_acceptor",
                to="eveonline.EveCharacter",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="issuer_corporation",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contract_issuer_corporation",
                to="eveonline.EveCorporationInfo",
            ),
        ),
        migrations.AlterField(
            model_name="eveorganization",
            name="category",
            field=models.CharField(
                choices=[
                    ("alliance", "Alliance"),
                    ("corporation", "Corporation"),
                    ("character", "Character"),
                ],
                max_length=32,
            ),
        ),
    ]



# Generated by Django 2.2.5 on 2019-10-29 18:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("freight", "0006_auto_20191028_2204"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="contracthandler",
            options={
                "verbose_name": "Contract Handler [my_alliance]",
                "verbose_name_plural": "Contract Handler [my_alliance]",
            },
        ),
        migrations.AddField(
            model_name="contracthandler",
            name="operation_mode",
            field=models.CharField(
                choices=[
                    ("my_alliance", "My Alliance"),
                    ("my_corporation", "My Corporation"),
                ],
                default="my_alliance",
                help_text="defines what kind of contracts are synced",
                max_length=32,
            ),
        ),
    ]
